#ifndef GIVENS_HPP
#define GIVENS_HPP
#include <utils/Exceptions.hpp>
#include <utils/PugsAssert.hpp>

#include <cmath>
#include <iomanip>

class Givens
{
 private:
  static void
  _givens(const double a, const double b, double& c, double& s)
  {
    if (b == 0) {
      c = 1;
      s = 0;
    } else {
      if (std::abs(b) > std::abs(a)) {
        const double tau = -a / b;
        s                = 1. / sqrt(1 + tau * tau);
        c                = s * tau;
      } else {
        Assert(a != 0);
        const double tau = -b / a;
        c                = 1 / sqrt(1 + tau * tau);
        s                = c * tau;
      }
    }
  }

  static void
  _rotate(const double Ai_1j, const double Aij, const double c, const double s, double& ui_1, double& ui)
  {
    ui_1 = c * Ai_1j - s * Aij;
    ui   = s * Ai_1j + c * Aij;
  }

  template <typename MatrixType, typename VectorType, typename RHSVectorType>
  static void
  _lift(const MatrixType& A, const RHSVectorType& b, VectorType& x)
  {
    for (ssize_t i = x.dimension() - 1; i >= 0; --i) {
      const double inv_Aii = 1 / A(i, i);
      double sum           = 0;
      for (size_t j = i + 1; j < x.dimension(); ++j) {
        sum += A(i, j) * x[j];
      }
      x[i] = (b[i] - sum) * inv_Aii;
    }
  }

  template <typename MatrixType, typename UnknownMatrixType, typename RHSMatrixType>
  static void
  _liftCollection(const MatrixType& A, const RHSMatrixType& B, UnknownMatrixType& X)
  {
    for (ssize_t i = X.numberOfRows() - 1; i >= 0; --i) {
      const double inv_Aii = 1 / A(i, i);
      for (size_t k = 0; k < X.numberOfColumns(); ++k) {
        double sum = 0;
        for (size_t j = i + 1; j < X.numberOfRows(); ++j) {
          sum += A(i, j) * X(j, k);
        }
        X(i, k) = (B(i, k) - sum) * inv_Aii;
      }
    }
  }

 public:
  template <typename MatrixType, typename VectorType, typename RHSVectorType>
  static void
  solveInPlace(MatrixType& A, VectorType& x, RHSVectorType& b)
  {
    Assert(x.dimension() == A.numberOfColumns(), "The number of columns of A must be the size of x");
    Assert(b.dimension() == A.numberOfRows(), "The number of rows of A must be the size of b");

    for (size_t j = 0; j < A.numberOfColumns(); ++j) {
      for (size_t i = A.numberOfRows() - 1; i > j; --i) {
        double c;
        double s;
        _givens(A(i - 1, j), A(i, j), c, s);
        for (size_t k = j; k < A.numberOfColumns(); ++k) {
          double xi_1(0), xi(0);
          _rotate(A(i - 1, k), A(i, k), c, s, xi_1, xi);
          A(i - 1, k) = xi_1;
          A(i, k)     = xi;
        }
        double xi_1(0), xi(0);
        _rotate(b[i - 1], b[i], c, s, xi_1, xi);
        b[i - 1] = xi_1;
        b[i]     = xi;
      }
    }

    _lift(A, b, x);
  }

  template <typename MatrixType, typename VectorType, typename RHSVectorType>
  static void
  solve(const MatrixType& A, VectorType& x, const RHSVectorType& b)
  {
    Assert(x.dimension() == A.numberOfColumns(), "The number of columns of A must be the size of x");
    Assert(b.dimension() == A.numberOfRows(), "The number of rows of A must be the size of b");

    MatrixType rotateA    = copy(A);
    RHSVectorType rotateb = copy(b);

    solveInPlace(rotateA, x, rotateb);
  }

  template <typename MatrixType, typename UnknownMatrixType, typename RHSMatrixType>
  static void
  solveCollectionInPlace(MatrixType& A, UnknownMatrixType& X, RHSMatrixType& B)
  {
    Assert(X.numberOfRows() == A.numberOfColumns(), "The number of columns of A must be the number of rows of X");
    Assert(B.numberOfRows() == A.numberOfRows(), "The number of rows of A must be the rows of B");
    Assert(X.numberOfColumns() == B.numberOfColumns(), "The number of columns of X and B must be the same");

    for (size_t j = 0; j < A.numberOfColumns(); ++j) {
      for (size_t i = A.numberOfRows() - 1; i > j; --i) {
        double c;
        double s;
        _givens(A(i - 1, j), A(i, j), c, s);
        for (size_t k = j; k < A.numberOfColumns(); ++k) {
          double xi_1(0), xi(0);
          _rotate(A(i - 1, k), A(i, k), c, s, xi_1, xi);
          A(i - 1, k) = xi_1;
          A(i, k)     = xi;
        }

        for (size_t l = 0; l < B.numberOfColumns(); ++l) {
          double xi_1(0), xi(0);
          _rotate(B(i - 1, l), B(i, l), c, s, xi_1, xi);
          B(i - 1, l) = xi_1;
          B(i, l)     = xi;
        }
      }
    }

    _liftCollection(A, B, X);
  }

  template <typename MatrixType, typename UnknownMatrixType, typename RHSMatrixType>
  static void
  solveCollection(const MatrixType& A, UnknownMatrixType& X, const RHSMatrixType& B)
  {
    Assert(X.numberOfRows() == A.numberOfColumns(), "The number of columns of A must be the number of rows of X");
    Assert(B.numberOfRows() == A.numberOfRows(), "The number of rows of A must be the rows of B");
    Assert(X.numberOfColumns() == B.numberOfColumns(), "The number of columns of X and B must be the same");

    MatrixType rotateA    = copy(A);
    RHSMatrixType rotateB = copy(B);

    solveCollectionInPlace(rotateA, X, rotateB);
  }

  Givens()  = delete;
  ~Givens() = delete;
};

#endif   // GIVENS_HPP
