#include <language/modules/LocalFSIModule.hpp>

#include <language/modules/SchemeModuleTypes.hpp>
#include <language/modules/MeshModuleTypes.hpp>

#include <language/modules/ModuleRepository.hpp>
#include <language/utils/BuiltinFunctionEmbedder.hpp>


#include <scheme/CouplingBoundaryConditionDescriptor.hpp>


#include <scheme/LocalDtAcousticSolver.hpp>
#include <scheme/LocalDtHyperelasticSolver.hpp>

#include <scheme/Order2AcousticSolver.hpp>
#include <scheme/Order2HyperelasticSolver.hpp>
#include <scheme/Order2LocalDtHyperelasticSolver.hpp>
#include <scheme/DiscreteFunctionUtils.hpp>


#include <memory>

LocalFSIModule::LocalFSIModule()
{

  this->_addBuiltinFunction("coupling",
                            std::function(

                              [](std::shared_ptr<const IBoundaryDescriptor> boundary,
                                 const size_t& label) -> std::shared_ptr<const IBoundaryConditionDescriptor> {
                                return std::make_shared<CouplingBoundaryConditionDescriptor>(boundary, label);
                              }

                              ));
                              

    this->_addBuiltinFunction("eucclhyd_solver_order2",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& c,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& gamma,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& pinf,
                                 const double& dt) -> std::tuple<std::shared_ptr<const MeshVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>> {
                                return Order2AcousticSolverHandler{getCommonMesh({rho, u, E, c, p})}
                                  .solver()
                                  .apply(Order2AcousticSolverHandler::SolverType::Eucclhyd, dt, rho, u, E, c, p,
                                         bc_descriptor_list, gamma, pinf);
                              }

                              ));

    this->_addBuiltinFunction("glace_solver_order2",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& c,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& gamma,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& pinf,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list,
                                 const double& dt) -> std::tuple<std::shared_ptr<const MeshVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>> {
                                return Order2AcousticSolverHandler{getCommonMesh({rho, u, E, c, p})}
                                  .solver()
                                  .apply(Order2AcousticSolverHandler::SolverType::Glace, dt, rho, u, E, c, p,
                                         bc_descriptor_list, gamma, pinf);
                              }

                              ));

  this->_addBuiltinFunction("local_dt_glace_solver",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& c1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p1,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list1,
                                 const double& dt1, const std::shared_ptr<const DiscreteFunctionVariant>& rho2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& c2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p2,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list2,
                                 const size_t& q)
                                -> std::tuple<
                                  std::shared_ptr<const MeshVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const MeshVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>> {
                                return LocalDtAcousticSolverHandler{getCommonMesh({rho1, u1, E1, c1, p1}),
                                                                    getCommonMesh({rho2, u2, E2, c2, p2})}
                                  .solver()
                                  .apply(LocalDtAcousticSolverHandler::SolverType::Glace, dt1, q, rho1, rho2, u1, u2,
                                         E1, E2, c1, c2, p1, p2, bc_descriptor_list1, bc_descriptor_list2);
                              }

                              ));

  this->_addBuiltinFunction("local_dt_glace_solver",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& c1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p1,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list1,
                                 const double& dt1, const std::shared_ptr<const DiscreteFunctionVariant>& rho2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& c2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p2,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list2)
                                -> std::tuple<
                                  std::shared_ptr<const MeshVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const MeshVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>> {
                                return LocalDtAcousticSolverHandler{getCommonMesh({rho1, u1, E1, c1, p1}),
                                                                    getCommonMesh({rho2, u2, E2, c2, p2})}
                                  .solver()
                                  .apply(LocalDtAcousticSolverHandler::SolverType::Glace, dt1, rho1, rho2, u1, u2, E1,
                                         E2, c1, c2, p1, p2, bc_descriptor_list1, bc_descriptor_list2);
                              }

                              ));

  this->_addBuiltinFunction("local_dt_eucclhyd_solver",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& c1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p1,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list1,
                                 const double& dt1, const std::shared_ptr<const DiscreteFunctionVariant>& rho2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& c2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p2,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list2,
                                 const size_t& q)
                                -> std::tuple<
                                  std::shared_ptr<const MeshVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const MeshVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>> {
                                return LocalDtAcousticSolverHandler{getCommonMesh({rho1, u1, E1, c1, p1}),
                                                                    getCommonMesh({rho2, u2, E2, c2, p2})}
                                  .solver()
                                  .apply(LocalDtAcousticSolverHandler::SolverType::Eucclhyd, dt1, q, rho1, rho2, u1, u2,
                                         E1, E2, c1, c2, p1, p2, bc_descriptor_list1, bc_descriptor_list2);
                              }

                              ));

  this->_addBuiltinFunction("local_dt_eucclhyd_solver",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& c1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p1,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list1,
                                 const double& dt1, const std::shared_ptr<const DiscreteFunctionVariant>& rho2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& c2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p2,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list2)
                                -> std::tuple<
                                  std::shared_ptr<const MeshVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const MeshVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>> {
                                return LocalDtAcousticSolverHandler{getCommonMesh({rho1, u1, E1, c1, p1}),
                                                                    getCommonMesh({rho2, u2, E2, c2, p2})}
                                  .solver()
                                  .apply(LocalDtAcousticSolverHandler::SolverType::Eucclhyd, dt1, rho1, rho2, u1, u2,
                                         E1, E2, c1, c2, p1, p2, bc_descriptor_list1, bc_descriptor_list2);
                              }

                              ));

    this->_addBuiltinFunction("hyperelastic_eucclhyd_neohook_solver_order2",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& lambda,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& mu,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& gamma,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p_inf,
                                 const double& dt) -> std::tuple<std::shared_ptr<const MeshVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>> {
                                return Order2HyperelasticSolverHandler{getCommonMesh({rho, u, E, CG, aL, aT, sigma})}
                                  .solver()
                                  .apply(Order2HyperelasticSolverHandler::SolverType::Eucclhyd, 1, dt, rho, u, E, CG, aL, aT,
                                         sigma, p, bc_descriptor_list, lambda, mu, gamma, p_inf);
                              }

                              ));

        this->_addBuiltinFunction("hyperelastic_eucclhyd_neohook_solver_order2",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& lambda,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& mu,
                                 const double& dt) -> std::tuple<std::shared_ptr<const MeshVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>> {
                                return Order2HyperelasticSolverHandler{getCommonMesh({rho, u, E, CG, aL, aT, sigma})}
                                  .solver()
                                  .apply(Order2HyperelasticSolverHandler::SolverType::Eucclhyd, 3, dt, rho, u, E, CG, aL, aT,
                                         sigma, p, bc_descriptor_list, lambda, mu, lambda, mu);
                              }

                              ));

        this->_addBuiltinFunction("hyperelastic_eucclhyd_neohook2_solver_order2",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& lambda,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& mu,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& gamma,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p_inf,
                                 const double& dt) -> std::tuple<std::shared_ptr<const MeshVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>> {
                                return Order2HyperelasticSolverHandler{getCommonMesh({rho, u, E, CG, aL, aT, sigma})}
                                  .solver()
                                  .apply(Order2HyperelasticSolverHandler::SolverType::Eucclhyd, 2, dt, rho, u, E, CG, aL, aT,
                                         sigma, p, bc_descriptor_list, lambda, mu, gamma, p_inf);
                              }

                              ));

  this->_addBuiltinFunction("local_dt_hyperelastic_eucclhyd_neohook_solver",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma1,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& rho2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma2,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& lambda2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& mu2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& gamma2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p_inf2,
                                 const double& dt1, 
                                 const size_t& q)
                                -> std::tuple<
                                  std::shared_ptr<const MeshVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const MeshVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>> {
                                return LocalDtHyperelasticSolverHandler{getCommonMesh(
                                                                          {rho1, u1, E1, CG1, aL1, aT1, sigma1}),
                                                                        getCommonMesh(
                                                                          {rho2, u2, E2, CG2, aL2, aT2, sigma2})}
                                  .solver()
                                  .apply(LocalDtHyperelasticSolverHandler::SolverType::Eucclhyd, 1, dt1, q, rho1, rho2, u1,
                                         u2, E1, E2, CG1, CG2, aL1, aL2, aT1, aT2, sigma1, sigma2, bc_descriptor_list1,
                                         bc_descriptor_list2, lambda2, mu2, gamma2, p_inf2);
                              }

                              ));

  this->_addBuiltinFunction("local_dt_hyperelastic_eucclhyd_neohook_solver",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma1,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& rho2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma2,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& lambda2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& mu2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& gamma2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p_inf2,
                                 const double& dt1)
                                -> std::tuple<
                                  std::shared_ptr<const MeshVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const MeshVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>> {
                                return LocalDtHyperelasticSolverHandler{getCommonMesh(
                                                                          {rho1, u1, E1, CG1, aL1, aT1, sigma1}),
                                                                        getCommonMesh(
                                                                          {rho2, u2, E2, CG2, aL2, aT2, sigma2})}
                                  .solver()
                                  .apply(LocalDtHyperelasticSolverHandler::SolverType::Eucclhyd, 1, dt1, rho1, rho2, u1,
                                         u2, E1, E2, CG1, CG2, aL1, aL2, aT1, aT2, sigma1, sigma2, bc_descriptor_list1,
                                         bc_descriptor_list2, lambda2, mu2, gamma2, p_inf2);
                              }

                              ));

  this->_addBuiltinFunction("local_dt_hyperelastic_eucclhyd_neohook2_solver",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma1,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& rho2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma2,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& lambda2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& mu2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& gamma2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p_inf2,
                                 const double& dt1, 
                                 const size_t& q)
                                -> std::tuple<
                                  std::shared_ptr<const MeshVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const MeshVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>> {
                                return LocalDtHyperelasticSolverHandler{getCommonMesh(
                                                                          {rho1, u1, E1, CG1, aL1, aT1, sigma1}),
                                                                        getCommonMesh(
                                                                          {rho2, u2, E2, CG2, aL2, aT2, sigma2})}
                                  .solver()
                                  .apply(LocalDtHyperelasticSolverHandler::SolverType::Eucclhyd, 2, dt1, q, rho1, rho2, u1,
                                         u2, E1, E2, CG1, CG2, aL1, aL2, aT1, aT2, sigma1, sigma2, bc_descriptor_list1,
                                         bc_descriptor_list2, lambda2, mu2, gamma2, p_inf2);
                              }

                              ));

  this->_addBuiltinFunction("local_dt_hyperelastic_eucclhyd_neohook2_solver",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma1,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& rho2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma2,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& lambda2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& mu2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& gamma2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p_inf2,
                                 const double& dt1)
                                -> std::tuple<
                                  std::shared_ptr<const MeshVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const MeshVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>> {
                                return LocalDtHyperelasticSolverHandler{getCommonMesh(
                                                                          {rho1, u1, E1, CG1, aL1, aT1, sigma1}),
                                                                        getCommonMesh(
                                                                          {rho2, u2, E2, CG2, aL2, aT2, sigma2})}
                                  .solver()
                                  .apply(LocalDtHyperelasticSolverHandler::SolverType::Eucclhyd, 2, dt1, rho1, rho2, u1,
                                         u2, E1, E2, CG1, CG2, aL1, aL2, aT1, aT2, sigma1, sigma2, bc_descriptor_list1,
                                         bc_descriptor_list2, lambda2, mu2, gamma2, p_inf2);
                              }

                              ));

  this->_addBuiltinFunction("local_dt_hyperelastic_eucclhyd_neohook_solver_order2",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p1,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& rho2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p2,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& lambda1, 
                                 const std::shared_ptr<const DiscreteFunctionVariant>& mu1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& lambda2, 
                                 const std::shared_ptr<const DiscreteFunctionVariant>& mu2, 
                                 const std::shared_ptr<const DiscreteFunctionVariant>& gamma1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p_inf1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& gamma2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p_inf2, 
                                 const double& dt1)
                                -> std::tuple<
                                  std::shared_ptr<const MeshVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const MeshVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>> {
                                return Order2LocalDtHyperelasticSolverHandler{getCommonMesh(
                                                                          {rho1, u1, E1, CG1, aL1, aT1, sigma1}),
                                                                        getCommonMesh(
                                                                          {rho2, u2, E2, CG2, aL2, aT2, sigma2})}
                                  .solver()
                                  .apply(Order2LocalDtHyperelasticSolverHandler::SolverType::Eucclhyd, 1, dt1, rho1, rho2, u1,
                                         u2, E1, E2, CG1, CG2, aL1, aL2, aT1, aT2, sigma1, sigma2, p1, p2, bc_descriptor_list1,
                                         bc_descriptor_list2, lambda1, mu1, lambda2, mu2, gamma1, p_inf1, gamma2, p_inf2);
                              }

                              ));

  this->_addBuiltinFunction("local_dt_hyperelastic_eucclhyd_neohook_solver_order2",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p1,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& rho2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p2,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& lambda1, 
                                 const std::shared_ptr<const DiscreteFunctionVariant>& mu1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& lambda2, 
                                 const std::shared_ptr<const DiscreteFunctionVariant>& mu2, 
                                 const std::shared_ptr<const DiscreteFunctionVariant>& gamma1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p_inf1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& gamma2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p_inf2, 
                                 const double& dt1,
                                 const size_t& q)
                                -> std::tuple<
                                  std::shared_ptr<const MeshVariant>, 
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>, 
                                  std::shared_ptr<const MeshVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>> {
                                return Order2LocalDtHyperelasticSolverHandler{getCommonMesh(
                                                                          {rho1, u1, E1, CG1, aL1, aT1, sigma1}),
                                                                        getCommonMesh(
                                                                          {rho2, u2, E2, CG2, aL2, aT2, sigma2})}
                                  .solver()
                                  .apply(Order2LocalDtHyperelasticSolverHandler::SolverType::Eucclhyd, 1, dt1, q, rho1, rho2, u1,
                                         u2, E1, E2, CG1, CG2, aL1, aL2, aT1, aT2, sigma1, sigma2, p1, p2, bc_descriptor_list1,
                                         bc_descriptor_list2, lambda1, mu1, lambda2, mu2, gamma1, p_inf1, gamma2, p_inf2);
                              }

                              ));


  this->_addBuiltinFunction("local_dt_hyperelastic_eucclhyd_neohook2_solver_order2",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p1,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& rho2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p2,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& lambda1, 
                                 const std::shared_ptr<const DiscreteFunctionVariant>& mu1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& lambda2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& mu2, 
                                 const std::shared_ptr<const DiscreteFunctionVariant>& gamma1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p_inf1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& gamma2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p_inf2, 
                                 const double& dt1)
                                -> std::tuple<
                                  std::shared_ptr<const MeshVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const MeshVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>> {
                                return Order2LocalDtHyperelasticSolverHandler{getCommonMesh(
                                                                          {rho1, u1, E1, CG1, aL1, aT1, sigma1}),
                                                                        getCommonMesh(
                                                                          {rho2, u2, E2, CG2, aL2, aT2, sigma2})}
                                  .solver()
                                  .apply(Order2LocalDtHyperelasticSolverHandler::SolverType::Eucclhyd, 2, dt1, rho1, rho2, u1,
                                         u2, E1, E2, CG1, CG2, aL1, aL2, aT1, aT2, sigma1, sigma2, p1, p2, bc_descriptor_list1,
                                         bc_descriptor_list2, lambda1, mu1, lambda2, mu2, gamma1, p_inf1, gamma2, p_inf2);
                              }

                              ));

  this->_addBuiltinFunction("local_dt_hyperelastic_eucclhyd_neohook2_solver_order2",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p1,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& rho2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p2,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& lambda1, 
                                 const std::shared_ptr<const DiscreteFunctionVariant>& mu1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& lambda2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& mu2, 
                                 const std::shared_ptr<const DiscreteFunctionVariant>& gamma1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p_inf1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& gamma2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p_inf2, 
                                 const double& dt1,
                                 const size_t& q)
                                -> std::tuple<
                                  std::shared_ptr<const MeshVariant>, 
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>, 
                                  std::shared_ptr<const MeshVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>> {
                                return Order2LocalDtHyperelasticSolverHandler{getCommonMesh(
                                                                          {rho1, u1, E1, CG1, aL1, aT1, sigma1}),
                                                                        getCommonMesh(
                                                                          {rho2, u2, E2, CG2, aL2, aT2, sigma2})}
                                  .solver()
                                  .apply(Order2LocalDtHyperelasticSolverHandler::SolverType::Eucclhyd, 2, dt1, q, rho1, rho2, u1,
                                         u2, E1, E2, CG1, CG2, aL1, aL2, aT1, aT2, sigma1, sigma2, p1, p2, bc_descriptor_list1,
                                         bc_descriptor_list2, lambda1, mu1, lambda2, mu2, gamma1, p_inf1, gamma2, p_inf2);
                              }

                              ));

  this->_addBuiltinFunction("local_dt_hyperelastic_glace_neohook_solver",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma1,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& rho2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma2,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& lambda2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& mu2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& gamma2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p_inf2,
                                 const double& dt1, 
                                 const size_t& q)
                                -> std::tuple<
                                  std::shared_ptr<const MeshVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const MeshVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>> {
                                return LocalDtHyperelasticSolverHandler{getCommonMesh(
                                                                          {rho1, u1, E1, CG1, aL1, aT1, sigma1}),
                                                                        getCommonMesh(
                                                                          {rho2, u2, E2, CG2, aL2, aT2, sigma2})}
                                  .solver()
                                  .apply(LocalDtHyperelasticSolverHandler::SolverType::Glace, 1, dt1, q, rho1, rho2, u1,
                                         u2, E1, E2, CG1, CG2, aL1, aL2, aT1, aT2, sigma1, sigma2, bc_descriptor_list1,
                                         bc_descriptor_list2, lambda2, mu2, gamma2, p_inf2);
                              }

                              ));

  this->_addBuiltinFunction("local_dt_hyperelastic_glace_neohook_solver",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma1,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& rho2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma2,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& lambda2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& mu2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& gamma2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p_inf2,
                                 const double& dt1)
                                -> std::tuple<
                                  std::shared_ptr<const MeshVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const MeshVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>> {
                                return LocalDtHyperelasticSolverHandler{getCommonMesh(
                                                                          {rho1, u1, E1, CG1, aL1, aT1, sigma1}),
                                                                        getCommonMesh(
                                                                          {rho2, u2, E2, CG2, aL2, aT2, sigma2})}
                                  .solver()
                                  .apply(LocalDtHyperelasticSolverHandler::SolverType::Glace, 1, dt1, rho1, rho2, u1,
                                         u2, E1, E2, CG1, CG2, aL1, aL2, aT1, aT2, sigma1, sigma2, bc_descriptor_list1,
                                         bc_descriptor_list2, lambda2, mu2, gamma2, p_inf2);
                              }

                              ));

  this->_addBuiltinFunction("local_dt_hyperelastic_glace_neohook2_solver",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma1,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& rho2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma2,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& lambda2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& mu2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& gamma2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p_inf2,
                                 const double& dt1, 
                                 const size_t& q)
                                -> std::tuple<
                                  std::shared_ptr<const MeshVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const MeshVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>> {
                                return LocalDtHyperelasticSolverHandler{getCommonMesh(
                                                                          {rho1, u1, E1, CG1, aL1, aT1, sigma1}),
                                                                        getCommonMesh(
                                                                          {rho2, u2, E2, CG2, aL2, aT2, sigma2})}
                                  .solver()
                                  .apply(LocalDtHyperelasticSolverHandler::SolverType::Glace, 2, dt1, q, rho1, rho2, u1,
                                         u2, E1, E2, CG1, CG2, aL1, aL2, aT1, aT2, sigma1, sigma2, bc_descriptor_list1,
                                         bc_descriptor_list2, lambda2, mu2, gamma2, p_inf2);
                              }

                              ));

  this->_addBuiltinFunction("local_dt_hyperelastic_glace_neohook2_solver",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma1,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list1,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& rho2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& CG2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aL2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& aT2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma2,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& lambda2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& mu2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& gamma2,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p_inf2,
                                 const double& dt1)
                                -> std::tuple<
                                  std::shared_ptr<const MeshVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const MeshVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>,
                                  std::shared_ptr<const DiscreteFunctionVariant>> {
                                return LocalDtHyperelasticSolverHandler{getCommonMesh(
                                                                          {rho1, u1, E1, CG1, aL1, aT1, sigma1}),
                                                                        getCommonMesh(
                                                                          {rho2, u2, E2, CG2, aL2, aT2, sigma2})}
                                  .solver()
                                  .apply(LocalDtHyperelasticSolverHandler::SolverType::Glace, 2, dt1, rho1, rho2, u1,
                                         u2, E1, E2, CG1, CG2, aL1, aL2, aT1, aT2, sigma1, sigma2, bc_descriptor_list1,
                                         bc_descriptor_list2, lambda2, mu2, gamma2, p_inf2);
                              }

                              ));
}

void
LocalFSIModule::registerOperators() const
{

}

void
LocalFSIModule::registerCheckpointResume() const
{

}

ModuleRepository::Subscribe<LocalFSIModule> localFSI_module;
