#ifndef LOCAL_DT_UTILS_HPP
#define LOCAL_DT_UTILS_HPP

#include <language/utils/InterpolateItemValue.hpp>
#include <mesh/ItemValueUtils.hpp>
#include <mesh/ItemValueVariant.hpp>
#include <mesh/MeshFaceBoundary.hpp>
#include <mesh/MeshFlatNodeBoundary.hpp>
#include <mesh/MeshNodeBoundary.hpp>
#include <mesh/MeshTraits.hpp>
#include <mesh/StencilArray.hpp>
#include <mesh/StencilManager.hpp>
#include <mesh/SubItemValuePerItemVariant.hpp>
#include <scheme/DirichletBoundaryConditionDescriptor.hpp>
#include <scheme/DiscreteFunctionDPk.hpp>
#include <scheme/DiscreteFunctionDPkVariant.hpp>
#include <scheme/DiscreteFunctionP0.hpp>
#include <scheme/DiscreteFunctionUtils.hpp>
#include <scheme/DiscreteFunctionVariant.hpp>
#include <scheme/ExternalBoundaryConditionDescriptor.hpp>
#include <scheme/FixedBoundaryConditionDescriptor.hpp>
#include <scheme/IBoundaryConditionDescriptor.hpp>
#include <scheme/IDiscreteFunctionDescriptor.hpp>
#include <scheme/PolynomialReconstruction.hpp>
#include <scheme/PolynomialReconstructionDescriptor.hpp>
#include <scheme/SymmetryBoundaryConditionDescriptor.hpp>
#include <utils/Socket.hpp>

template <MeshConcept MeshType>
std::tuple<const std::shared_ptr<const DiscreteFunctionVariant>,
             const std::shared_ptr<const DiscreteFunctionVariant>,
             const std::shared_ptr<const DiscreteFunctionVariant>,
             const std::shared_ptr<const DiscreteFunctionVariant>>
  apply_state_law(const size_t& law,
                  const std::shared_ptr<const DiscreteFunctionVariant>& rho_v,
                  const std::shared_ptr<const DiscreteFunctionVariant>& u_v,
                  const std::shared_ptr<const DiscreteFunctionVariant>& E_v,
                  const std::shared_ptr<const DiscreteFunctionVariant>& CG_v,
                  const std::shared_ptr<const DiscreteFunctionVariant>& lambda_v,
                  const std::shared_ptr<const DiscreteFunctionVariant>& mu_v,
                  const std::shared_ptr<const DiscreteFunctionVariant>& gamma_v,
                  const std::shared_ptr<const DiscreteFunctionVariant>& p_inf_v)
  {
    std::shared_ptr mesh_v = getCommonMesh({rho_v, u_v});

    const DiscreteFunctionP0<const double>& rho_d                         = rho_v->get<DiscreteFunctionP0<const double>>();
    const DiscreteFunctionP0<const TinyVector<MeshType::Dimension>>& u_d  = u_v->get<DiscreteFunctionP0<const TinyVector<MeshType::Dimension>>>();
    const DiscreteFunctionP0<const double>& E_d                           = E_v->get<DiscreteFunctionP0<const double>>();
    const DiscreteFunctionP0<const TinyMatrix<MeshType::Dimension>>& CG_d = CG_v->get<DiscreteFunctionP0<const TinyMatrix<MeshType::Dimension>>>();
    const DiscreteFunctionP0<const double>& lambda_d                      = lambda_v->get<DiscreteFunctionP0<const double>>();
    const DiscreteFunctionP0<const double>& mu_d                          = mu_v->get<DiscreteFunctionP0<const double>>();
    const DiscreteFunctionP0<const double>& gamma_d                       = gamma_v->get<DiscreteFunctionP0<const double>>();
    const DiscreteFunctionP0<const double>& p_inf_d                       = p_inf_v->get<DiscreteFunctionP0<const double>>();

    const TinyMatrix<MeshType::Dimension> I = identity;
    const DiscreteFunctionP0<const double>& eps_d = E_d - 0.5 * dot(u_d, u_d);

    DiscreteFunctionP0<double> p_d(mesh_v);
    DiscreteFunctionP0<TinyMatrix<MeshType::Dimension>> sigma_d(mesh_v);
    DiscreteFunctionP0<double> aL_d(mesh_v);
    DiscreteFunctionP0<double> aT_d(mesh_v);

    if (law == 1) {
      p_d = (gamma_d - 1) * rho_d * eps_d - p_inf_d * gamma_d;
      sigma_d = (mu_d / sqrt(det(CG_d)) * (CG_d - I) + lambda_d / sqrt(det(CG_d)) * log(sqrt(det(CG_d))) * I) - p_d * I;
      aL_d = sqrt((lambda_d + 2 * mu_d) / rho_d + gamma_d * (p_d + p_inf_d) / rho_d);
      aT_d = sqrt(mu_d / rho_d);
    } else if (law == 2){
      p_d = (gamma_d - 1) * rho_d * eps_d - gamma_d * p_inf_d;
      sigma_d = 2*mu_d / sqrt(det(CG_d)) * (CG_d - 1. / 3. * trace(CG_d) * I) - p_d * I;
      aL_d = sqrt((2 * mu_d) / rho_d + (gamma_d * (p_d + p_inf_d)) / rho_d);
      aT_d = sqrt(mu_d / rho_d);
    } else {
      sigma_d = (mu_d / sqrt(det(CG_d)) * (CG_d - I) + lambda_d / sqrt(det(CG_d)) * log(sqrt(det(CG_d))) * I);
      p_d = -1./3. * trace(sigma_d);
      aL_d = sqrt((lambda_d + 2 * mu_d) / rho_d);
      aT_d = sqrt(mu_d / rho_d);
    }
    return {std::make_shared<DiscreteFunctionVariant>(sigma_d), 
            std::make_shared<DiscreteFunctionVariant>(p_d),
            std::make_shared<DiscreteFunctionVariant>(aL_d),
            std::make_shared<DiscreteFunctionVariant>(aT_d)};
  }

  template<MeshConcept MeshType>
  std::tuple<std::shared_ptr<const MeshVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>>
  mesh_correction(const std::shared_ptr<const MeshVariant>& i_mesh1,
                  const std::shared_ptr<const MeshVariant>& i_mesh2,
                  const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                  const std::shared_ptr<const DiscreteFunctionVariant>& u,
                  const std::shared_ptr<const DiscreteFunctionVariant>& E,
                  const std::shared_ptr<const DiscreteFunctionVariant>& CG,
                  std::vector<NodeId>& map1,
                  std::vector<NodeId>& map2)
  {
    const MeshType& mesh1 = *i_mesh1->get<MeshType>();
    const MeshType& mesh2 = *i_mesh2->get<MeshType>();

    NodeValue<TinyVector<MeshType::Dimension>> xr1     = copy(mesh1.xr());
    NodeValue<TinyVector<MeshType::Dimension>> new_xr2 = copy(mesh2.xr());

    size_t n = map1.size();

    for (size_t i = 0; i < n; i++) {
      for (size_t j = 0; j < MeshType::Dimension; j++) {
        new_xr2[map2[i]][j] = xr1[map1[i]][j];
      }
    }

    std::shared_ptr<const MeshType> new_mesh2 = std::make_shared<MeshType>(mesh2.shared_connectivity(), new_xr2);
    const std::shared_ptr<const MeshVariant>& mesh_copy = std::make_shared<MeshVariant>(new_mesh2);

    const std::shared_ptr<const DiscreteFunctionVariant>& new_rho = shallowCopy(mesh_copy, rho);
    const std::shared_ptr<const DiscreteFunctionVariant>& new_u   = shallowCopy(mesh_copy, u);
    const std::shared_ptr<const DiscreteFunctionVariant>& new_E   = shallowCopy(mesh_copy, E);
    const std::shared_ptr<const DiscreteFunctionVariant>& new_CG  = shallowCopy(mesh_copy, CG);

    return std::make_tuple(mesh_copy,new_rho,new_u,new_E,new_CG);
  }

#endif //LOCAL_DT_UTILS_HPP