#ifndef ORDER2_LIMITERS_HPP
#define ORDER2_LIMITERS_HPP

#include <algebra/EigenvalueSolver.hpp>

#include <language/utils/InterpolateItemValue.hpp>
#include <mesh/ItemValueUtils.hpp>
#include <mesh/ItemValueVariant.hpp>
#include <mesh/MeshFaceBoundary.hpp>
#include <mesh/MeshFlatNodeBoundary.hpp>
#include <mesh/MeshNodeBoundary.hpp>
#include <mesh/MeshTraits.hpp>
#include <mesh/StencilArray.hpp>
#include <mesh/StencilManager.hpp>
#include <mesh/SubItemValuePerItemVariant.hpp>
#include <scheme/DirichletBoundaryConditionDescriptor.hpp>
#include <scheme/DiscreteFunctionDPk.hpp>
#include <scheme/DiscreteFunctionDPkVariant.hpp>
#include <scheme/DiscreteFunctionP0.hpp>
#include <scheme/DiscreteFunctionUtils.hpp>
#include <scheme/DiscreteFunctionVariant.hpp>
#include <scheme/ExternalBoundaryConditionDescriptor.hpp>
#include <scheme/FixedBoundaryConditionDescriptor.hpp>
#include <scheme/IBoundaryConditionDescriptor.hpp>
#include <scheme/IDiscreteFunctionDescriptor.hpp>
#include <scheme/PolynomialReconstruction.hpp>
#include <scheme/PolynomialReconstructionDescriptor.hpp>
#include <scheme/SymmetryBoundaryConditionDescriptor.hpp>
#include <utils/Socket.hpp>

template <MeshConcept MeshType>
void
scalar_limiter(const MeshType& mesh,
               const DiscreteFunctionP0<const double>& f,
               DiscreteFunctionDPk<MeshType::Dimension, double>& DPk_fh)
{
    MeshData<MeshType>& mesh_data = MeshDataManager::instance().getMeshData(mesh);
    StencilManager::BoundaryDescriptorList symmetry_boundary_descriptor_list;
    StencilDescriptor stencil_descriptor{1, StencilDescriptor::ConnectionType::by_nodes};
    auto stencil = StencilManager::instance().getCellToCellStencilArray(mesh.connectivity(), stencil_descriptor, symmetry_boundary_descriptor_list);

    const auto xj = mesh_data.xj();

    CellValue<double> alph_J2{mesh.connectivity()};

    parallel_for(mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id){
      const double fj = f[cell_id];

      double f_min = fj;
      double f_max = fj;

      const auto cell_stencil = stencil[cell_id];
      for(size_t i_cell = 0; i_cell < cell_stencil.size(); ++i_cell){
        f_min = std::min(f_min, f[cell_stencil[i_cell]]);
        f_max = std::max(f_max, f[cell_stencil[i_cell]]);
      }

      double f_bar_min = fj;
      double f_bar_max = fj;

      for(size_t i_cell = 0; i_cell < cell_stencil.size(); ++i_cell){
        const CellId cell_k_id = cell_stencil[i_cell];
        const double f_xk    = DPk_fh[cell_id](xj[cell_k_id]);

        f_bar_min = std::min(f_bar_min, f_xk);
        f_bar_max = std::max(f_bar_max, f_xk);
      }

      const double eps = 1E-14;
      double coef1     = 1;
      if (std::abs(f_bar_max - fj) > eps) {
        coef1 = (f_max - fj) / ((f_bar_max - fj));
      }

      double coef2 = 1.;
      if (std::abs(f_bar_min - fj) > eps) {
        coef2 = (f_min - fj) / ((f_bar_min - fj));
      }

      const double lambda = std::max(0., std::min(1., std::min(coef1, coef2)));
      alph_J2[cell_id] = lambda;

      auto coefficients = DPk_fh.coefficients(cell_id);

      coefficients[0] = (1 - lambda) * f[cell_id] + lambda * coefficients[0];
      for (size_t i = 1; i < coefficients.size(); ++i) {
        coefficients[i] *= lambda;
      }
    });
  }

  template <MeshConcept MeshType>
  void
  vector_limiter(const MeshType& mesh,
                 const DiscreteFunctionP0<const TinyVector<MeshType::Dimension>>& f,
                 DiscreteFunctionDPk<MeshType::Dimension, TinyVector<MeshType::Dimension>> DPk_fh,
                 const CellValue<const TinyMatrix<MeshType::Dimension>>& grad_u) 
  {
    MeshData<MeshType>& mesh_data = MeshDataManager::instance().getMeshData(mesh);
    StencilManager::BoundaryDescriptorList symmetry_boundary_descriptor_list;
    StencilDescriptor stencil_descriptor{1, StencilDescriptor::ConnectionType::by_nodes};
    auto stencil = StencilManager::instance().getCellToCellStencilArray(mesh.connectivity(), stencil_descriptor,
                                                                        symmetry_boundary_descriptor_list);
    
    const auto xj = mesh_data.xj();

    CellValue<TinyMatrix<MeshType::Dimension>> sym_grad_u{mesh.connectivity()};
    parallel_for(
      mesh.numberOfCells(), PUGS_LAMBDA(const CellId j){
        sym_grad_u[j] = (grad_u[j] + transpose(grad_u[j]));
    });

    CellValue<SmallArray<double>> eigunvalues{mesh.connectivity()};
    CellValue<std::vector<SmallVector<double>>> eigunvectors{mesh.connectivity()};
    for(CellId j = 0; j < mesh.numberOfCells(); ++j){
      EigenvalueSolver{}.computeForSymmetricMatrix(sym_grad_u[j],eigunvalues[j],eigunvectors[j]);
    }

    CellValue<TinyVector<MeshType::Dimension>> n{mesh.connectivity()};
    CellValue<TinyVector<MeshType::Dimension>> t{mesh.connectivity()};
    CellValue<TinyVector<MeshType::Dimension>> l{mesh.connectivity()};
    parallel_for(
      mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
        n[cell_id] = zero;
        t[cell_id] = zero;
        l[cell_id] = zero;
        //if(frobeniusNorm(sym_grad_u[cell_id]) > 1E-5){
          if constexpr(MeshType::Dimension == 2){
            TinyVector<MeshType::Dimension> nj;
            TinyVector<MeshType::Dimension> tj;
            for(size_t i = 0; i < MeshType::Dimension; ++i){
              nj[i] = eigunvectors[cell_id][0][i];
              tj[i] = eigunvectors[cell_id][1][i];
            }
            n[cell_id] = nj;
            t[cell_id] = tj;
            if(l2Norm(nj) != 0){
              n[cell_id] = (1/l2Norm(nj))*n[cell_id];
            }
            if(l2Norm(tj) != 0){
              t[cell_id] = (1/l2Norm(tj))*t[cell_id];
            }
          } else {
            if constexpr(MeshType::Dimension == 3){
              TinyVector<MeshType::Dimension> nj;
              TinyVector<MeshType::Dimension> tj;
              TinyVector<MeshType::Dimension> lj;
              for(size_t i = 0; i < MeshType::Dimension; ++i){
                nj[i] = eigunvectors[cell_id][0][i];
                tj[i] = eigunvectors[cell_id][1][i];
                lj[i] = eigunvectors[cell_id][2][i];
              }
              n[cell_id] = nj;
              t[cell_id] = tj;
              l[cell_id] = lj;
              if(l2Norm(nj) != 0){
                n[cell_id] = (1/l2Norm(nj))*n[cell_id];
              }
              if(l2Norm(tj) != 0){
                t[cell_id] = (1/l2Norm(tj))*t[cell_id];
              }
              if(l2Norm(lj) != 0){
                l[cell_id] = (1/l2Norm(lj))*l[cell_id];
              }
            } else {
                TinyVector<MeshType::Dimension> nj;
                nj[0] = 1.;
                n[cell_id] = nj;
            }
          }
        //}
    });
    
    parallel_for(
      mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id){
        const double fn = dot(f[cell_id],n[cell_id]);
        const double ft = dot(f[cell_id],t[cell_id]);
        const double fl = dot(f[cell_id],l[cell_id]);

        double fn_min = fn;
        double fn_max = fn;
        double ft_min = ft;
        double ft_max = ft;
        double fl_min = fl;
        double fl_max = fl;

        const auto cell_stencil = stencil[cell_id];
        for(size_t i_cell = 0; i_cell < cell_stencil.size(); ++i_cell){
          const double fn_k = dot(f[cell_stencil[i_cell]],n[cell_id]);
          fn_min            = std::min(fn_min,fn_k);
          fn_max            = std::max(fn_max,fn_k);

          const double ft_k = dot(f[cell_stencil[i_cell]],t[cell_id]);
          ft_min            = std::min(ft_min,ft_k);
          ft_max            = std::max(ft_max,ft_k);

          const double fl_k = dot(f[cell_stencil[i_cell]],l[cell_id]);
          fl_min            = std::min(fl_min,fl_k);
          fl_max            = std::max(fl_max,fl_k);
        }

        double fn_bar_min = fn;
        double fn_bar_max = fn;
        double ft_bar_min = ft;
        double ft_bar_max = ft;
        double fl_bar_min = fl;
        double fl_bar_max = fl;

        for(size_t i_cell = 0; i_cell < cell_stencil.size(); ++i_cell){
          const CellId cell_k_id = cell_stencil[i_cell];
          const TinyVector<MeshType::Dimension> f_xk          = DPk_fh[cell_id](xj[cell_k_id]);

          const double fn_xk = dot(f_xk,n[cell_id]);
          fn_bar_min         = std::min(fn_bar_min,fn_xk);
          fn_bar_max         = std::max(fn_bar_max,fn_xk);

          const double ft_xk = dot(f_xk,t[cell_id]);
          ft_bar_min         = std::min(ft_bar_min,ft_xk);
          ft_bar_max         = std::max(ft_bar_max,ft_xk);

          const double fl_xk = dot(f_xk,l[cell_id]);
          fl_bar_min         = std::min(fl_bar_min,fl_xk);
          fl_bar_max         = std::max(fl_bar_max,fl_xk);
        }

        const double eps = 1E-14;
        double coef1_n   = 1;
        if(std::abs(fn_bar_max - fn) > eps){
          coef1_n = (fn_max - fn) / (fn_bar_max - fn);
        } 
        double coef2_n = 1;
        if(std::abs(fn_bar_min - fn) > eps){
          coef2_n = (fn_min - fn) / (fn_bar_min - fn);
        }
        
        double coef1_t = 1;
        if(std::abs(ft_bar_max - ft) > eps){
          coef1_t = (ft_max - ft) / (ft_bar_max - ft);
        }
        double coef2_t = 1;
        if(std::abs(ft_bar_min - ft) > eps){
          coef2_t = (ft_min - ft) / (ft_bar_min - ft);
        }

        double coef1_l = 1;
        if(std::abs(fl_bar_max - fl) > eps){
          coef1_l = (fl_max - fl) / (fl_bar_max - fl);
        }
        double coef2_l = 1;
        if(std::abs(fl_bar_min - fl) > eps){
          coef2_l = (fl_min - fl) / (fl_bar_min - fl);
        }

        const double lambda_n = std::max(0., std::min(1., std::min(coef1_n,coef2_n)));
        const double lambda_t = std::max(0., std::min(1., std::min(coef1_t,coef2_t)));
        const double lambda_l = std::max(0., std::min(1., std::min(coef1_l,coef2_l)));

        auto coefficients = DPk_fh.coefficients(cell_id);
        coefficients[0] = (1 - lambda_n)*fn*n[cell_id] + lambda_n*dot(coefficients[0],n[cell_id])*n[cell_id] 
                        + (1 - lambda_t)*ft*t[cell_id] + lambda_t*dot(coefficients[0],t[cell_id])*t[cell_id] 
                        + (1 - lambda_l)*fl*l[cell_id] + lambda_l*dot(coefficients[0],l[cell_id])*l[cell_id];
        for(size_t i = 1; i < coefficients.size(); ++i){
          coefficients[i] = lambda_n*dot(coefficients[i],n[cell_id])*n[cell_id] 
                          + lambda_t*dot(coefficients[i],t[cell_id])*t[cell_id] 
                          + lambda_l*dot(coefficients[i],l[cell_id])*l[cell_id];
        }
    });
  }

  template <MeshConcept MeshType>
  void
  matrix_limiter(const MeshType& mesh,
                 const DiscreteFunctionP0<const TinyMatrix<MeshType::Dimension>>& S,
                 DiscreteFunctionDPk<MeshType::Dimension, TinyMatrix<MeshType::Dimension>>& DPk_Sh)
  {
    MeshData<MeshType>& mesh_data = MeshDataManager::instance().getMeshData(mesh);
    StencilManager::BoundaryDescriptorList symmetry_boundary_descriptor_list;
    StencilDescriptor stencil_descriptor{1, StencilDescriptor::ConnectionType::by_nodes};
    auto stencil = StencilManager::instance().getCellToCellStencilArray(mesh.connectivity(), stencil_descriptor,
                                                                        symmetry_boundary_descriptor_list);

    const auto xj = mesh_data.xj();

    const DiscreteFunctionP0<const double>& inv2 = 0.5*trace(transpose(S)*S);
    parallel_for(
      mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id){
        const double inv2j = inv2[cell_id];
        
        double inv2_min = inv2j;
        double inv2_max = inv2j;

        const auto cell_stencil = stencil[cell_id];
        for(size_t i_cell = 0; i_cell < cell_stencil.size(); ++i_cell){
          inv2_min = std::min(inv2_min, inv2[cell_stencil[i_cell]]);
          inv2_max = std::max(inv2_max, inv2[cell_stencil[i_cell]]);
        }

        double inv2_bar_min = inv2j;
        double inv2_bar_max = inv2j;

        for(size_t i_cell = 0; i_cell < cell_stencil.size(); ++i_cell){
          const CellId cell_k_id = cell_stencil[i_cell];
          const double inv2_xk   = 0.5*trace(transpose(DPk_Sh[cell_id](xj[cell_k_id]))*DPk_Sh[cell_id](xj[cell_k_id]));

          inv2_bar_min = std::min(inv2_bar_min, inv2_xk);
          inv2_bar_max = std::max(inv2_bar_max, inv2_xk);
        }

        const double eps =  1E-14;
        double coef1     = 1.;
        double coef2     = 1.;
        if(std::abs(inv2_bar_max - inv2j) > eps){
          coef1 = (inv2_max - inv2j) / (inv2_bar_max - inv2j);
        } 
        if(std::abs(inv2_bar_min - inv2j) > eps){
          coef2 = (inv2_min - inv2j) / (inv2_bar_min - inv2j);
        }

        const double lambda_inv2 = sqrt(std::max(0., std::min(1., std::min(coef1, coef2))));

        auto coefficients = DPk_Sh.coefficients(cell_id);

        coefficients[0] = (1-lambda_inv2)*S[cell_id] + lambda_inv2 * coefficients[0];
        for (size_t i = 1; i < coefficients.size(); ++i) {
          coefficients[i] *= lambda_inv2;
        } 
    });
  } 

#endif //ORDER2_LIMITERS_HPP